import logging


class QuestionClassifier:
    
    def __init__(self,resources_dir):
        self.logger = logging.getLogger('muheqa')
        self.logger.debug("initializing Question Classifier ...")
            
    def get_category(self,question):
        # category: boolean, literal or resource
        # type: date, number or string
        result_dict = {
            'question': question,
            'category': 'literal',
            'type': 'string'
        }       
        return result_dict
            


import logging
import requests
from transformers import pipeline

class Retriever:
	
	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Retriever ...")

		self.logger.debug("loading eqa language models ...")
		#qa_language_model = "deepset/roberta-base-squad2-covid" #roberta-covid
		#qa_language_model = "deepset/roberta-base-squad2" #roberta
		#qa_language_model = "shaina/covid_qa_mpnet" #CoQUAD
		qa_language_models = ["deepset/roberta-base-squad2","shaina/covid_qa_mpnet"]
		self.qa_pipes = []	
		for lm in qa_language_models:
			self.qa_pipes.append(pipeline("question-answering", model=lm, tokenizer=lm))
		
		
	def get_evidences(self,question,context):
		evidences = []
		values = []
		for pipe in self.qa_pipes:
			evidence = {}
			try:
				if (len(context) == 0):
					return evidence
				result = pipe(question=question, context=context, min_answer_len=1, max_answer_len=100, padding=True, truncation=True)
				score = round(result['score'], 1)
				value = result['answer']
				if (value not in values):
					values.append(value)			
					evidence['value']=result['answer']
					evidence['score']=score
					evidence['summary']=context
					evidence['start']=result['start']
					evidence['end']=result['end']			
					evidences.append(evidence)					
			except Exception as e:
				print("Error extracting evidence:",e)
		return evidences



import logging
import requests
import muheqa.summary.texts.verbalizer as vb
import muheqa.summary.resources.wikipedia as kg_wikipedia
import muheqa.summary.resources.dbpedia as kg_dbpedia
import muheqa.summary.resources.d4c as db_d4c

class Splitter:
	
	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Splitter ...")
		
	def get_documents(self,sentences,max_lenght=427):
		counter = 0
		doc_sentences = []
		documents = []
		for s in sentences:
			num_tokens = len(s.split(" "))	
			doc_sentences.append(s)
			counter += num_tokens
			if (counter > max_lenght):
				document = " . ".join(doc_sentences)
				documents.append(document)
				doc_sentences = []
				counter = 0			
		if (len(doc_sentences)>0):
			document = " . ".join(doc_sentences)
			documents.append(document)
		return documents


import logging
import requests
import muheqa.summary.texts.verbalizer as vb
import muheqa.summary.resources.wikipedia as kg_wikipedia
import muheqa.summary.resources.dbpedia as kg_dbpedia
import muheqa.summary.resources.d4c as db_d4c

class Ranking:
	
	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Ranking ...")

	def get_top_evidences(self,evidences,n=1):
		self.logger.debug("top " + str(n) + " evidences from " + str(len(evidences)) + " evidences")
		self.logger.debug(str(evidences))
		if (len(evidences)<2):
			return evidences
		evidences.sort(key=lambda x: x.get('score'),reverse=True)
		best_score = evidences[0]['score']
		top_evidences = []
		for index, c in enumerate(evidences):
			if (n < 0) or (index < n) or (c['score'] == best_score):
				#if (c['score'] > 0):
				top_evidences.append(c)
		return top_evidences
		



import os
import json
import muheqa.connector as mhqa

kb = mhqa.connect(dbpedia=True)

# assign directory
input_directory = '/Users/cbadenes/Library/CloudStorage/Dropbox/Trabajo/Carlos/Academic/DoctoradoIA/estancias/2023/Cardiff/plan_trabajo/dataset/questions'
output_directory = '/Users/cbadenes/Library/CloudStorage/Dropbox/Trabajo/Carlos/Academic/DoctoradoIA/estancias/2023/Cardiff/plan_trabajo/dataset/answers'

def answer_evidences(news):
    for annotation in news['annotation_values']:
        for value in annotation['5w1h_value']:
            question = value['kb_question_en']
            print(question)
            response = kb.query(question, max_answers=3)
            print(response)
            #d4c_response = d4c.query(question, max_answers=3)
            #print("D4C:", d4c_response)
            #kbs_response = kbs.query(question, max_answers=3)
            #value['kbs_response'] = kbs_response
    return news  



# iterate over files in that directory
counter = 1
for filename in os.listdir(input_directory):
    input_path = os.path.join(input_directory, filename)
    output_path = os.path.join(output_directory, filename)
    print("[",counter,"]",input_path)
    counter += 1
    # checking if it is a file
    if os.path.isfile(output_path):
        continue
    if os.path.isfile(input_path):                
        with open(input_path, "r") as jsonFile:
            data = json.load(jsonFile)
        annotated_news = answer_evidences(data)         


import logging
import requests
import muheqa.evidence.documents.splitter as sp
import muheqa.evidence.responses.retriever as rt
import muheqa.evidence.candidates.ranking as rk

class Discoverer:
	
	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Discoverer ...")
		
		self.splitter 	= sp.Splitter()
		self.retriever 	= rt.Retriever()
		self.ranking 	= rk.Ranking()


	def get_evidences(self, question, summary, max=3):
		max_tokens = 25
		docs = self.splitter.get_documents(summary,max_tokens)
		evidences = []
		for d in docs:
			for evidence in self.retriever.get_evidences(question, d):
				evidences.append(evidence)
		top_evidences = self.ranking.get_top_evidences(evidences,max)
		return top_evidences



import logging

# uncased NER model
from transformers import AutoTokenizer, AutoModelForTokenClassification
from transformers import pipeline

class Entity:

	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Entity class instance...")

		self.logger.debug("loading NER model ...")
		ner_language_model = "dslim/bert-base-NER-uncased"
		ner_tokenizer = AutoTokenizer.from_pretrained(ner_language_model)
		ner_model = AutoModelForTokenClassification.from_pretrained(ner_language_model)
		self.ner_nlp = pipeline("ner", model=ner_model, tokenizer=ner_tokenizer)

	def get(self,text):
		self.logger.debug("getting entities...")
		entities = []
		entity = ""
		index = -1
		offset = -1
		for token in self.ner_nlp(text):
			if (index == -1):
				index = token['index']
				offset = token['start']
			word = token['word']
			if (word[0] == '#'):
				word = token['word'].replace("#","")
			if (token['start']== offset):
				entity += word
			elif (token['index']-index < 2):
				entity += " " + word
			else:
				entities.append(entity)
				entity = word		
			index = token['index']
			offset = token['end']
		if (len(entity) > 0):    
			entities.append(entity)
		return entities



	


import logging
import requests
import muheqa.summary.keywords.concept as cp
import muheqa.summary.keywords.discovery as kw
import muheqa.summary.texts.verbalizer as vb
import muheqa.summary.resources.wikipedia as kg_wikipedia
import muheqa.summary.resources.dbpedia as kg_dbpedia
import muheqa.summary.resources.d4c as db_d4c

class Summarizer:
	
	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Summarizer ...")
	
		self.concept 	= cp.Concept()
		self.discovery 	= kw.Discovery()		
		self.verbalizer = vb.Verbalizer()		
		self.wikipedia 	= kg_wikipedia.Wikipedia()
		self.dbpedia 	= kg_dbpedia.DBpedia()
		self.d4c 		= db_d4c.D4C()



	def get_sentences(self,query,max_resources=5,wikipedia=True,dbpedia=True,d4c=True,by_name=True,by_properties=True,by_description=True):
		key = query + str(max_resources) + str(wikipedia) + str(dbpedia) + str(d4c) + str(by_name) + str(by_properties) + str(by_description)

		# Create Summary
		sentences = []

		## Keywords to search KG Resources
		keywords = self.discovery.get_keywords(query)
		if ((len(keywords['entities']) == 0) and (len(keywords['concepts']) == 0)):
			self.logger.warn("no keywords found in question")
			return sentences

		self.logger.debug("keywords: " + str(keywords))
		keys = keywords['entities']
		if (len(keys) == 0):
			keys =  keywords['concepts']
		for kw in keys:
			if (wikipedia):
				wiki_sentences = self.verbalizer.kg_to_text(self.wikipedia,query,kw,max_resources,by_name,by_properties,by_description)
				self.logger.debug("wiki sentences:" + str(wiki_sentences))
				sentences.extend(wiki_sentences)

			if (dbpedia):
				dbpedia_sentences = self.verbalizer.kg_to_text(self.dbpedia,query,kw,max_resources,by_name,by_properties,by_description)
				self.logger.debug("dbpedia sentences:" + str(dbpedia_sentences))
				sentences.extend(dbpedia_sentences)

		## Concepts to search texts
		if (d4c):

			self.logger.debug("Concepts: " + str(keywords['concepts']))
			
			rows = 3
			if (max_resources > 1):
				rows = max_resources*2
			d4c_sentences = self.verbalizer.db_to_text(self.d4c, query, keywords['entities'], keywords['concepts'], rows)
			self.logger.debug("d4c sentences:" + str(len(d4c_sentences)))
			sentences.extend(d4c_sentences)			

		return sentences


import logging

import warnings
warnings.filterwarnings("ignore", category=FutureWarning)

class CustomFormatter(logging.Formatter):

	grey = "\x1b[38;20m"
	blue = "\x1b[33;94m"
	yellow = "\x1b[33;20m"
	red = "\x1b[31;20m"
	bold_red = "\x1b[31;1m"
	reset = "\x1b[0m"
	format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

	FORMATS = {
		logging.DEBUG: grey + format + reset,
		logging.INFO: blue + format + reset,
		logging.WARNING: yellow + format + reset,
		logging.ERROR: red + format + reset,
		logging.CRITICAL: bold_red + format + reset
	}

	def format(self, record):
		log_fmt = self.FORMATS.get(record.levelno)
		formatter = logging.Formatter(log_fmt)
		return formatter.format(record)



import logging

# PoS tagger
from flair.data import Sentence
from flair.models import SequenceTagger

class Concept:

	def __init__(self):
		logging.getLogger("flair").setLevel(level=logging.WARNING)
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Concept class instance...")

		pos_language_model = "flair/pos-english"
		self.ner_tagger = SequenceTagger.load(pos_language_model)
		
	def get(self,text):
		self.logger.debug("getting concepts ...")
		main_categories=['CD','NN','NNS','NNP','NNPS']
		additional_categories=['JJ','NNS']
		drop_categories=['IN','WRB','CC']

		# make sentence
		sentence = Sentence(text)

		# predict NER tags
		self.ner_tagger.predict(sentence)

		# iterate over tokens
		concepts = []
		current_concept = ""
		partial_concept = ""
		previous_token = ""
		for t in sentence.tokens:
			for label in t.annotation_layers.keys():
				token = t.text
				category = t.get_labels(label)[0].value
				self.logger.debug("Token:"+ token + " [ " + category + "]  Previous: [" + previous_token + "]")
				#self.logger.debug("Current Concept: " + current_concept + ", Partial Concept: " + partial_concept)
				if (category in main_categories):
					if (len(partial_concept) > 0 ):
						current_concept = partial_concept + " " + token
						partial_concept = ""
					elif (current_concept == ""):
						current_concept += token
					else:
						current_concept += " " + token
				elif (category in additional_categories):
					if (len(current_concept)>0):
						current_concept += " " + token
					elif (len(partial_concept)>0):
						partial_concept += " " + token
					elif (previous_token not in ['WRB','CC']):
						partial_concept += token
					else:
						if (len(current_concept)>0):
							concepts.append(current_concept)
						current_concept = ""
						partial_concept = ""
				elif(category in drop_categories):
					if (len(current_concept)>0):
							concepts.append(current_concept)
					current_concept = ""
					partial_concept = ""
				elif len(current_concept) > 0:
					if (len(current_concept)>0):
							concepts.append(current_concept)
					current_concept = ""
					partial_concept = ""
			previous_token = t.get_labels(label)[0].value
		if (len(current_concept) > 0):
			concepts.append(current_concept)
		return concepts




	



import logging
import requests
import nltk
nltk.download('omw-1.4')
nltk.download('wordnet')
from nltk.stem import WordNetLemmatizer
from sentence_transformers import SentenceTransformer
from sklearn.metrics.pairwise import cosine_similarity
import unidecode


class Graph:

	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Graph Explorer...")
		self.logger.debug("loading sentence-transformer...")
		self.sentence_model = SentenceTransformer("sentence-transformers/all-distilroberta-v1")
		self.logger.debug("loading wordnet lemmatizer...")
		self.lemmatizer = WordNetLemmatizer()

	def lemmatize(self, text):
	    result = []
	    for token in text.split(" "):
	    	result.append(self.lemmatizer.lemmatize(token))
	    return " ".join(result)

	def sort_by_similarity(self,ref_text,texts):	
		sentences = [unidecode.unidecode(ref_text.strip())]
		sentences.extend([unidecode.unidecode(t.strip()) for t in texts])
		embeddings = self.sentence_model.encode(sentences)
		sim_list = []
		index=0
		for e in embeddings[1:]:
			ref = embeddings[0]
			score = cosine_similarity([ref], [e])
			score_val = round(score[0][0], 1)
			sim_list.append({'id':index, 'text':texts[index], 'score':score_val})  
			index+=1
		sim_list.sort(key=lambda x: x.get('score'),reverse=True)
		return sim_list

	def get_top_similar(self,ref_text,resources,max=-1):
		top_resources = []
		if (len(resources) == 0):
			return top_resources
		sorted_resources = self.sort_by_similarity(ref_text,[c['text'] for c in resources])  
		best_score = sorted_resources[0]['score']
		for index, c in enumerate(sorted_resources):
			if (max < 0) or (index < max) or (c['score'] == best_score):
				candidate = resources[c['id']]
				candidate['score'] = c['score']
				top_resources.append(candidate)
		return top_resources

	def get_top_resources(self,context,label,candidates,max=-1,by_name=True,by_properties=True,by_description=True):
		if (len(candidates) == 0):
			return []
		
		# initialize score		
		sorted_candidates = []	
		for c in candidates:
			candidate = c
			candidate['score'] = 1.0
			sorted_candidates.append(candidate)


		# filter by similar resource name and label
		if (by_name):
			names = [ {'id':i, 'text':c['label'] } for i,c in enumerate(candidates)]
			top_candidates_by_name = self.get_top_similar(label,names,max)
			sorted_candidates = []
			for t in top_candidates_by_name:
				candidate = candidates[t['id']]
				# normalize score
				candidate['score'] = t['score']
				sorted_candidates.append(candidate)
  		
		# filter by similar resource propery and context
		if(by_properties):
			properties = []
			for i,c in enumerate(sorted_candidates):
				for p in c['properties']:
					properties.append({'id':i, 'text':p['value'] })
			top_candidates_by_prop = self.get_top_similar(context.replace(label,""),properties,-1) 
			new_sorted_candidates = []
			for t in top_candidates_by_prop:
				candidate = sorted_candidates[t['id']]
				# only the best property is considered
				if (candidate not in new_sorted_candidates):
					# normalize score
					candidate['score'] = (2*candidate['score'] + 4*t['score']) / 6.0
					new_sorted_candidates.append(candidate)
			sorted_candidates = new_sorted_candidates[:max]

		# filter by similar resource description and context
		if(by_description):
			descriptions = [ {'id':i, 'text':c['description'] } for i,c in enumerate(sorted_candidates)]
			top_candidates_by_desc = self.get_top_similar(context.replace(label,""),descriptions,max)
			new_sorted_candidates = []    
			for t in top_candidates_by_desc:
				candidate = sorted_candidates[t['id']]
				# only the best description is considered
				if (candidate not in new_sorted_candidates):
					# normalize score
					candidate['score'] = (2*candidate['score'] + 1*t['score']) / 3.0
					new_sorted_candidates.append(candidate)
			sorted_candidates = new_sorted_candidates
		return sorted_candidates
	
		


import logging
import muheqa.summary.resources.graph as kg_graph
import muheqa.summary.resources.wikipedia as kg_wikipedia
import muheqa.summary.resources.dbpedia as kg_dbpedia
import muheqa.summary.resources.d4c as db_d4c


class Verbalizer:

	def __init__(self):
		self.logger 	= logging.getLogger('muheqa')
		self.logger.debug("initializing Verbalizer ...")
		self.graph 		= kg_graph.Graph()
		
	def property_to_text(self,s,p,o):
		if (len(o)==0):
			return ""
		property_value = o
		if (len(o)>0):
			property_value = ", ".join(o)		
		tokens = ["The",p,"of",s,"is",property_value]   
		return " ".join(tokens)
	

	def kg_to_text(self,kg,query,keyword,max_resources=5,by_name=True,by_properties=True,by_description=True):
		sentences = []
		
		# find related resources
		kg_resources = kg.find_resources(keyword)	

		# sort resources by name, properties and description
		top_resources = self.graph.get_top_resources(query,keyword,kg_resources,max_resources,by_name,by_properties,by_description)
		
		# get top properties for each resource
		for resource in top_resources:
			self.logger.debug("Top Resource: '" + str(resource['id'])+ "'")
			if ('properties' in resource):
				properties = [{ 'id':r['id'] , 'text':r['value']} for r in resource['properties']]
				top_properties = self.graph.get_top_similar(query.replace(keyword,""),properties, 5)
				self.logger.debug("top properties" + str(top_properties) + " from resource: " + str(resource['id']))
				for p in top_properties:
					self.logger.debug("verbalizing property" + str(p))
					fw_values = kg.get_forward_property_value(resource['id'],p['id'])
					if (len(fw_values)>0):
						t = self.property_to_text(resource['label'],p['text'],[o['value'] for o in fw_values])
						sentences.append(t)						

					#bw_values = kg.get_backward_property_value(resource['id'],p['id'])	
					#if (len(bw_values)>0):
					#	t = self.property_to_text(resource['label'],p['text'],[o['value'] for o in bw_values])
					#	sentences.append(t)
		return sentences

	def db_to_text(self, db, query, keywords, concepts, max_texts=5):
		db_texts = db.find_texts(query,keywords, concepts, max_texts)
		self.logger.debug("DB Texts: " + str(len(db_texts)))
		return db_texts




import logging
import muheqa.answer.analysis.classifier as cl
import muheqa.answer.generation.model as mo

class Composer:
	
	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Composer ...")
		
		self.model 			= mo.ModelEN()
		self.classifier 	= cl.QuestionClassifier("./resources_dir")


	def get_answers(self, question, evidences, max=3):
		category = self.classifier.get_category(question)
		self.logger.debug("Question Analysis: " + str(category))
		answers = []
		for e in evidences:
			answer = self.model.get_response(category,e)
			answers.append(answer)
		return answers[:max]



import logging
import requests
import unidecode
from SPARQLWrapper import SPARQLWrapper, JSON
import muheqa.summary.resources.graph as kg_graph


class DBpedia:

    def __init__(self):
        self.logger = logging.getLogger('muheqa')
        self.logger.debug("initializing DBpedia retriever...")
        self.sparql = SPARQLWrapper(
            "https://dbpedia.org/sparql/", agent='Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36')
        self.sparql.setReturnFormat(JSON)
        self.sparql.setTimeout(timeout=60)
        self.graph 	= kg_graph.Graph()

    def get_property_value(self, filter):
        query = """
			PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
			PREFIX dbr: <http://dbpedia.org/resource/> 
			select distinct ?object ?label {
			{ FILTER }
			optional { 
			?object rdfs:label ?label .
			filter langMatches(lang(?label), 'en')
			}
			}
			LIMIT 150
		"""
        query_text = query.replace('FILTER', filter)
        self.sparql.setQuery(query_text)
        result = []
        while (len(result) == 0):
            try:
                ret = self.sparql.queryAndConvert()
                for r in ret["results"]["bindings"]:
                    id = r['object']['value']
                    value = id
                    if ('label' in r) and ('value' in r['label']):
                        value = r['label']['value']
                    if (' id ' not in value.lower()) and (' link ' not in value.lower()) and ('has abstract' not in value.lower()) and ('wiki' not in value.lower()) and ('instance of' not in value.lower()):
                        result.append(
                            {'id': id, 'value': unidecode.unidecode(value).replace("\n*","")})
            except Exception as e:
                print("Error on wikidata property value query:",
                      e, "->", query_text)
            break
        return result

    def get_forward_property_value(self, entity, property):
        query_filter = "<http://dbpedia.org/resource/ENTITY> <PROPERTY> ?object"
        self.logger.debug("fw property:" + str(property) + " of entity: " + str(entity))
        return self.get_property_value(query_filter.replace("ENTITY", entity).replace("PROPERTY", property))

    def get_backward_property_value(self, entity, property):
        query_filter = "?object <PROPERTY> <http://dbpedia.org/resource/ENTITY>"
        self.logger.debug("bk property:" + str(property) + " of entity: " + str(entity))
        return self.get_property_value(query_filter.replace("ENTITY", entity).replace("PROPERTY", property))

    def get_properties(self, entity):
        query = """
	  		PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
	  		PREFIX dbr: <http://dbpedia.org/resource/> 
	  		select distinct ?property ?label {
	  			{ <http://dbpedia.org/resource/ENTITY> ?property ?o }
	  			union
	  			{ ?s ?property <http://dbpedia.org/resource/ENTITY> }
	  			optional { 
	  			?property rdfs:label ?label .
	  			filter langMatches(lang(?label), 'en')
	  		}
	  		filter(regex(?property, "property", "i" )) 
  			}
  			LIMIT 150
  			"""
        query_text = query.replace('ENTITY', entity)
        self.sparql.setQuery(query_text)
        result = []
        try:
            ret = self.sparql.queryAndConvert()
            for r in ret["results"]["bindings"]:
                if ('label' in r) and ('value' in r['label']):
                    value = r['label']['value']
                    id = r['property']['value']
                    if ('id' not in value.lower()) and ('link' not in value.lower()) and ('has abstract' not in value.lower()) and ('wiki' not in value.lower()) and ('instance of' not in value.lower()):
                        result.append({'id': id, 'value': value})
        except Exception as e:
            print("Error on dbpedia property query:", e, "->", query_text)
        return result

    def find_resources(self, label):
        self.logger.debug("getting summary from DBpedia for label: " + label)
        candidates = []
        if (label == ""):
            return candidates     
        headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36'}
        query_path = "https://lookup.dbpedia.org/api/search?format=JSON&query=QUERY_TEXT&maxResults=10"
        request = query_path.replace("QUERY_TEXT", label)
        try:
            r = requests.get(request, headers=headers)
        except:
            self.logger.error("An error getting data from DBpedia lookup service: ")
            return []
        if (len(r.json()['docs']) == 0):
            lemma = self.graph.lemmatize(label)	
            self.logger.debug("search retried by lemma:" + str(lemma))
            r = requests.get(query_path.replace("QUERY_TEXT", lemma))
            size = len(label.split(" "))
            index = 1
            while(('search' in r.json()) and (len(r.json()['search']) == 0) and (index < size)):
                query_label = " ".join(label.split(" ")[index:])
                index += 1
                r = requests.get(query_path.replace("QUERY_TEXT", query_label))
                self.logger.debug("retried with: " + query_label)
        for answer in r.json()['docs']:
            description, label, id = "", "", ""
            properties = []
            if ('comment' in answer) and (len(answer['comment']) > 0):
                description = answer['comment'][0].replace("<B>", "").replace("</B>", "")
            if ('resource' in answer) and (len(answer['resource']) > 0):
                id = answer['resource'][0].split("http://dbpedia.org/resource/")[1]
                properties = self.get_properties(id)
            if ('label' in answer) and (len(answer['label']) > 0):
                label = answer['label'][0].replace("<B>", "").replace("</B>", "")
            else:
                label = id

            candidate = {
                'label': label,
                'id': id,
                'description': description,
                'properties': properties
            }
            self.logger.debug("DBpedia resource candidate: '" + str(id) + "'")
            candidates.append(candidate)
        return candidates



import logging
from urllib.request import urlopen
import urllib.parse
import json
import unidecode
import muheqa.summary.keywords.concept as cc
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM


class D4C:

	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing D4C retriever...")
		self.url = "http://librairy.linkeddata.es/solr/cord19-paragraphs"
		#self.url = "http://localhost:8983/solr/documents"
		self.concept_discovery = cc.Concept()
		
		model_name = "castorini/monot5-base-msmarco-10k" 
		self.tokenizer = AutoTokenizer.from_pretrained(model_name)
		self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name)




	def find_texts(self,query,entities, concepts, max=5):		
		q = ""
		if (len(entities)>0):
			q += "(" + " OR ".join([ "text_t:\""+t + "\"" for t in entities]) + ")"
		
		if (len(concepts)>0):
			unique_concepts = [ t for t in concepts if t not in entities]
			if (len(unique_concepts)>0):
				if (len(q)>0):
					q += " AND "
				q += "(" + " OR ".join([ "text_t:\""+t + "\"" for t in unique_concepts ]) +")"

		self.logger.debug("query => " + q)		
		try:
			connection = urlopen(self.url + '/select?fl=text_t,id&q='+urllib.parse.quote(q)+'&rows=50'+'&wt=json')
			response = json.load(connection)
		except:
			self.logger.error("Error getting data from " + self.url)
			return []
		self.logger.debug( str(len(response['response']['docs'])) + "/" + str(response['response']['numFound']) + " documents found.")
		docs = response['response']['docs']
		if (response['response']['numFound'] == 0) and (len(concepts)>0):
			try:
				tokens = []
				for c in concepts:
					for t in c.split(" "):
						tokens.append(t)
				q = "(" + " OR ".join([ "text_t:\""+t + "\"" for t in tokens]) +")"
				self.logger.debug("query => " + q)
				connection = urlopen(self.url + '/select?fl=text_t,id&q='+urllib.parse.quote(q)+'&rows=50'+'&wt=json')
				response = json.load(connection)
				self.logger.debug( str(len(response['response']['docs'])) + "/" + str(response['response']['numFound']) + " documents found.")
				docs = response['response']['docs']
			except e:				
				self.logger.error("error getting documents: " + str(e))
		sentences = []
		for document in docs:
			
			doc_text = document['text_t']
			is_valid = False
			if (len(doc_text) > 0 ) and (len(doc_text.split(" "))<512):
				#validate based on T5 instead of BM25 or TFIDF according to paper "Document Ranking with a Pretrained Sequence-to-Sequence Model" https://aclanthology.org/2020.findings-emnlp.63.pdf
				# castorini/monot5-base-msmarco-10k
				# castorini/monot5-3b-msmarco-10k
				try:
					input_ids = self.tokenizer(query + " " + doc_text, return_tensors="pt").input_ids  # Batch size 1
					outputs = self.model.generate(input_ids,max_new_tokens=1)
					is_valid = self.tokenizer.decode(outputs[0], skip_special_tokens=True) == "true"
				except e:
					self.logger.error("error in tokenizer: " + str(e))
			else:
				self.logger.warn("document length greater than max tokenizer")								
			#self.logger.debug(str(is_valid))
			if (is_valid):			
				sentences.append(doc_text)	
		return sentences




import logging

class ModelEN:

    def __init__(self):
        self.logger = logging.getLogger('muheqa')
        self.logger.debug("initializing Answer Model ...")        

    def contains_number(self,text):
        return any(char.isdigit() for char in text)    

    def get_response(self, category, evidence):
        response = {}
        if (category['category'] == 'resource'):
            response['type'] = 'literal'
            response['answer'] = evidence['value']
        elif (category['category'] == 'boolean'):
            response['type'] = 'boolean'
            response['answer'] = (evidence['score'] > 0.5)
        elif (len(category['type']) > 0) and (category['type'][0] == 'number'):
            response['type'] = 'number'
            if (self.contains_number(evidence['value'])):
                response['answer'] = evidence['value']
            elif (',' in evidence['value']):
                response['answer'] = len(evidence['value'].split(","))
            else:
                response['answer'] = 1
        else:
           response['answer'] = evidence['value'] 
        response['confidence'] = evidence['score']
        response['evidence'] = evidence['summary']
        response['start'] = evidence['start']
        response['end'] = evidence['end']
        return response


import logging
import requests
import muheqa.summary.resources.graph as kg_graph

from SPARQLWrapper import SPARQLWrapper, JSON


class Wikipedia:

	def __init__(self):
		self.logger 	= logging.getLogger('muheqa')
		self.logger.debug("initializing Wikipedia retriever...")
		self.sparql 	= SPARQLWrapper("https://query.wikidata.org/sparql",agent='Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36')
		self.sparql.setReturnFormat(JSON)
		self.sparql.setTimeout(timeout=60)
		self.graph 		= kg_graph.Graph()

	def get_property_value(self,filter):		
		query = """
		PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
		PREFIX bd: <http://www.bigdata.com/rdf#>
		PREFIX wd: <http://www.wikidata.org/entity/> 
		PREFIX wdt: <http://www.wikidata.org/prop/direct/>
		PREFIX wikibase: <http://wikiba.se/ontology#>
		SELECT distinct ?obj ?objLabel
		WHERE
		FILTER 
		LIMIT 150
		"""
		query_text = query.replace('FILTER',filter)
		self.sparql.setQuery(query_text)
		result = []
		while (len(result) == 0):
			try:
				ret = self.sparql.queryAndConvert() 
				for r in ret["results"]["bindings"]:
					id = r['obj']['value']
					value = id
					if ('objLabel' in r) and ('value' in r['objLabel']):
						value = r['objLabel']['value']                
					if (' id ' not in value.lower()) and (' link ' not in value.lower()) and ('has abstract' not in value.lower()) and ('wiki' not in value.lower()) and ('instance of' not in value.lower()):
						result.append({'id':id, 'value':value})
			except Exception as e:
				print("Error on wikidata property value query:",e,"->",query_text)
			break
		return result	


	def get_forward_property_value(self,entity,property): 
		query_filter = """
			{
				wd:ENTITY wdt:PROPERTY ?obj .
				SERVICE wikibase:label { bd:serviceParam wikibase:language "en". }
			}                            
			"""
		return self.get_property_value(query_filter.replace("ENTITY",entity).replace("PROPERTY",property))

	def get_backward_property_value(self,entity,property):
		query_filter = """
			{
				?obj wdt:PROPERTY wd:ENTITY .
				SERVICE wikibase:label { bd:serviceParam wikibase:language "en". }
			}                            
			"""
		return self.get_property_value(query_filter.replace("ENTITY",entity).replace("PROPERTY",property))	


	def get_properties(self, entity):
	  query = """
	      PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
	      PREFIX wd: <http://www.wikidata.org/entity/> 
	      SELECT distinct ?prop ?propLabel
	      WHERE
	      {
	        { wd:ENTITY ?a ?b }
	              union
	              { ?s ?a wd:ENTITY } .

	        SERVICE wikibase:label { bd:serviceParam wikibase:language "en". } 
	        ?prop wikibase:directClaim ?a .
	      } 
	      LIMIT 150
	      """
	  query_text = query.replace('ENTITY',entity)
	  self.sparql.setQuery(query_text)
	  result = []
	  try:
	        ret = self.sparql.queryAndConvert()
	        for r in ret["results"]["bindings"]:
	            if ('propLabel' in r) and ('value' in r['propLabel']):
	                    value = r['propLabel']['value']
	                    id = r['prop']['value'].split("http://www.wikidata.org/entity/")[1]
	                    if ('id' not in value.lower()) and ('link' not in value.lower()) and ('has abstract' not in value.lower()) and ('wiki' not in value.lower()) and ('instance of' not in value.lower()):
	                        result.append({'id':id, 'value':value})
	  except Exception as e:
	        print("Error on wikidata property query:",e,"->",query_text)           
	  return result

	def find_resources(self, label):
		self.logger.debug("getting summary from Wikipedia for label: " + label)
		if (label==""):
			return candidates
		candidates = []
		headers = { 'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36'}
		query_path = "https://www.wikidata.org/w/api.php?action=wbsearchentities&search=QUERY_TEXT&language=en&limit=10&type=item&format=json"
		request = query_path.replace("QUERY_TEXT",label)
		r = requests.get(request,headers = headers)
		if (len(r.json()['search']) == 0):
			lemma = self.graph.lemmatize(label)	
			self.logger.debug("retry search by lemma:" + str(lemma))
			r = requests.get(query_path.replace("QUERY_TEXT",lemma))
			size = len(label.split(" "))
			index = 1
			while(('search' in r.json()) and (len(r.json()['search']) == 0) and (index<size)):
				query_label = " ".join(label.split(" ")[index:])
				index += 1  
				self.logger.debug("retry search by Partial Label:" + query_label)
				r = requests.get(query_path.replace("QUERY_TEXT",query_label)) 
		for answer in r.json()['search']:
			description = ""
			if ('description' in answer['display']):
				description = answer['display']['description']['value']
				if 'disambiguation' in description:
					continue
			candidate = {
				'label': answer['display']['label']['value'],
				'id':answer['id'],
				'description' : description,
				'properties' : self.get_properties(answer['id'])
				}
			candidates.append(candidate)    
		return candidates  




import logging

import muheqa.summary.keywords.concept as cp
import muheqa.summary.keywords.entity as ent
import unidecode

class Discovery:

	def __init__(self):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing Keyword Discovery class instance...")

		self.entity_discovery = ent.Entity()
		self.concept_discovery = cp.Concept()

	def unicase(self,label):
  		return unidecode.unidecode(label.strip()).lower()

	def get_keywords(self,text):
		concepts = [self.unicase(e) for e in self.concept_discovery.get(text)]
		result = { 'entities':[], 'concepts':concepts}
		entities = self.entity_discovery.get(text)
		for e in entities:
			extended_entity = e
			for c in concepts:
				if (e in c):
					extended_entity = c
			if (extended_entity not in result['entities']):
				result['entities'].append(extended_entity)			
		return result


import logging
import muheqa.summary.summarizer as sm
import muheqa.evidence.discoverer as ds
import muheqa.answer.composer as cp


class Connection:

	def __init__(self, wikidata=True,dbpedia=False,d4c=False):
		self.logger = logging.getLogger('muheqa')
		self.logger.debug("initializing MuHeQA ...")
		self.summarizer = sm.Summarizer()
		self.discoverer = ds.Discoverer()
		self.composer   = cp.Composer()
		self.wikidata  = wikidata
		self.dbpedia    = dbpedia
		self.d4c        = d4c 

	def query(self,question,max_answers=1,max_resources=3,by_name=True,by_properties=True,by_description=True):
		self.logger.debug("Question: " + question)
		sentences = self.summarizer.get_sentences(question,max_resources,self.wikidata,self.dbpedia,self.d4c,by_name,by_properties,by_description)
		if (len(sentences) == 0):
			self.logger.warn("no summary created")
			return sentences
		evidences = self.discoverer.get_evidences(question,sentences,max_answers)
		self.logger.debug("Evidences: " + str(evidences))
		answers   = self.composer.get_answers(question,evidences,max_answers)
		self.logger.debug("Answers: " + str(answers))
		return answers

